#include "knaps.hpp"

#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <math.h>



GA oKS;                                         // Globale Variable fuer den GA



// Werte des GA an Tcl zurueckgeben:
//  Rueckgaben an Tcl: 1. Aktuelle Generation
//                     2. Anzahl der angefangenen Container
//                     3. Bester Fitness-Wert
//                     4. Mittlerer Fitness-Wert
//                     5. Schlechtester Fitness-Wert
//                     6. Summe der Fitness-Werte
//                     danach fuer jeden Block (s. GAInit, 2. Parameter):
//                      x-Koordinate y-Koordinate Breite Hoehe
void GAReturn(Tcl_Interp *interp)
{
    INDIVIDUAL Ind;

    Ind=oKS.oldpop().ind()[oKS.oldpop().best()];
    oKS.GetFitness(Ind.chrom());                // Bloecke fuer Besten anordnen
    sprintf(interp->result," %d %ld %ld %ld %ld %ld",
        oKS.generation().cv_nat0(),
        oKS.containers().cv_nat0(),
	(long)oKS.oldpop().bestf().cv_real(),
        (long)(oKS.oldpop().sumfitness().cv_real()/oKS.stringlen().cv_nat1()),
        (long)oKS.oldpop().worstf().cv_real,
	(long)oKS.oldpop().sumfitness().cv_real());
    for(int i=1;i<=oKS.stringlen().cv_nat1();++i)   // Alle Bloecke ausgeben
	sprintf(interp->result+strlen(interp->result)," %d %d %d %d",
            oKS.bl()[i].x().cv_nat0(),oKS.bl()[i].y().cv_nat0(),
            oKS.bl()[i].bwidth().cv_nat0(),oKS.bl()[i].bheight().cv_nat0());
}

// Zusaetzliche Funktion der neuen Shell zum initialisieren des GA
//  Tcl-Parameter:     1. Populationsgroesse
//                     2. Anzahl der Bloecke
//                     3. Container-Breite
//                     4. Container-Hoehe
//                     5. Kreuzungswahrscheinlichkeit
//                     6. Mutationswahrscheinlichkeit
//                     7. Maximale Aenderung bei PMX 
//                     8.  danach fuer jeden Block (s. 2.):
//                         Block-Breite  Block-Hoehe
//  Rueckgaben an Tcl: siehe GAReturn
int GAInit(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    CONTAINER con;                              // Container
    BLOCKS bl(empty);                           // Bloecke
    double dCross,dMut,dMaxChg;
    int i,iPop,iLen,cw,ch;
    char *argv8;

    if(argc!=9)                                  // Korrekte Parameteranzahl
    {
        sprintf(interp->result,"Nicht genuegend Parameter fuer 'GAInit': %d",argc);
        return TCL_ERROR;
    }
    else
    {
        sscanf(argv[1],"%d",&iPop);             // Populationsgroesse und
        sscanf(argv[2],"%d",&iLen);             //  Anzahl der Bloecke einlesen

        sscanf(argv[3],"%d",&cw);               // Container einlesen und "bauen"
        sscanf(argv[4],"%d",&ch);
        con=CONTAINER(cw,ch);

        sscanf(argv[5],"%lf",&dCross);          // Kreuzungwahrscheinlichkeit,
        sscanf(argv[6],"%lf",&dMut);            //  Mutations    "",
        sscanf(argv[7],"%lf",&dMaxChg);         //  und Max. Aenderung einlesen

        for(i=1,argv8=argv[8];i<=iLen;++i)      // Bloecke einlesen
        {
	    int w,h;

            sscanf(argv8,"%d",&w);              // Breite einlesen
            while(*++argv8!=' ')                // Zum naechsten Parameter
                if(argv8-argv[8]>strlen(argv[8]))
                {
                    sprintf(interp->result,"Falsche Blockanzahl bei 'GAInit'");
                    return TCL_ERROR;
                }   
            sscanf(argv8,"%d",&h);              // Hoehe einlesen
            while(i<iLen && *++argv8!=' ')      // Zum naechsten Parameter
                if(argv8-argv[8]>strlen(argv[8]))
                {
                    sprintf(interp->result,"Falsche Blockanzahl bei 'GAInit'");
                    return TCL_ERROR;
                }
            bl.appr(BLOCK(w,h,0,0));        // Neuen Block hinzufuegen
        }

        oKS.Init(con,bl,iPop,iLen,dCross,dMut,dMaxChg); // Eigentliche Initialisierung

        GAReturn(interp);

        return TCL_OK;                      // Keine Fehler
    }
}

// Zusaetzliche Funktion der neuen Shell zur Ermittlung der naechsten
// Generation:
// Es werden <iGen> Generationen erzeugt und das Ergebnis jeder Generation
// mit an Tcl zurueckgegeben:
//  Tcl-Parameter:     1. Generationen  (optional, Default ist 1)
//  Rueckgaben an Tcl: siehe GAReturn
int NextGen(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    int iGen=0,i;

    if(argc==2)                                 // Anzahl der Generationen als Parameter
	sscanf(argv[1],"%d",&iGen);
    if(!iGen) iGen=1;                           // Mindestens eine Genration

    while(iGen--)
        oKS.NextGen();                          // Naechste Generation berechnen

    GAReturn(interp);
    
    return TCL_OK;
}

// Zufallsfunktion:
//  Tcl-Parameter:     1. Maximalwert
//  Rueckgaben an Tcl: 1. Ganzzahlige Zufallszahl im Bereich 1-Maximalwert 
int Random(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    if(argc==2)
    {
        int rndmax;

        sscanf(argv[1],"%d",&rndmax);
        sprintf(interp->result,"%d",rand()%rndmax+1);
	return TCL_OK;
    }
    else
    {
        interp->result="Falsche Parameteranzahl fuer 'Random'";

        return TCL_ERROR;
    }
}


// Applikationsspezifische Initialisierungen:
// siehe auch: TkAppInit.c
// Kommandos "GAInit" "NextGen" "Random" werden zu den Tcl/Tk-Kommandos
// hinzugefuegt.
int Tcl_AppInit(Tcl_Interp *interp)
{
    Tk_Window main=Tk_MainWindow(interp);

    if(Tcl_Init(interp)==TCL_ERROR || Tk_Init (interp)==TCL_ERROR)
        return TCL_ERROR;
    else
    {                                           // Eigene Funktionen bekanntgeben
	Tcl_CreateCommand(interp,"GAInit",GAInit,(ClientData)NULL,
                          (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand(interp,"NextGen",NextGen,(ClientData)NULL,
                          (Tcl_CmdDeleteProc *)NULL);
	Tcl_CreateCommand(interp,"Random",Random,(ClientData)NULL,
                          (Tcl_CmdDeleteProc *)NULL);

#if TCL_MAJOR_VERSION > 7 || ( TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION >=4 )
    Tcl_SetVar(interp,"tcl_rcFileName","~/.wishrc",TCL_GLOBAL_ONLY);
#else
    tcl_RcFileName="~/.wishrc";
#endif
    }

	return TCL_OK;
}



#if TCL_MAJOR_VERSION > 7 || ( TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION >=4 )

int main(int argc, char * argv[]) {
    Tk_Main(argc, argv, Tcl_AppInit);
    return 0;
}

#else

// Verbindung zur 'main'-Prozedur der Tk-Library
EXTERN int main _ANSI_ARGS_((int argc,char *argv[]));
int (*tclDummyMainPtr)()=(int (*)())main;

#endif


// 'matherr' aus Tcl-Library muss genutzt werden
#ifdef NEED_MATHERR
extern int matherr();
int *tclDummyMathPtr=(int *)matherr;
#endif
